/*
 * Copyright 2003 Draagon Software LLC. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software LLC.
 * Use is subject to license terms.
 */

package com.draagon.meta;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
//import java.util.Iterator;

public abstract class MetaClassLoader extends MetaData
{
    private final static Log log = LogFactory.getLog( MetaClassLoader.class );

    private final static List<MetaClassLoader> metaClassLoaders;

    private final Map<String, MetaClass> metaClassCache;

    static {
      metaClassLoaders = Collections.synchronizedList( new ArrayList<MetaClassLoader>() );
    }

    public MetaClassLoader()
    {
      metaClassCache = Collections.synchronizedMap( new WeakHashMap<String, MetaClass>() );
      registerLoader( this );
    }
    
    public void init() throws Exception {
      log.info( "Loading the [" + getClass().getSimpleName() + "] MetaClassLoader" );
    }
    
    /**
     * Gets the primary MetaData class
     */
    public final Class<?> getMetaDataClass()
    {
      return MetaClassLoader.class;
    }

    /**
     * Retrieves the attribute options expected for this class; empty in this case
     */
    public Collection<AttributeOption> getAttributeOptions()
    {
      return new ArrayList<AttributeOption>();
    }

    /**
     * Retrieves the MetaClassLoader with the specified Name
     */
    public static Collection<MetaClassLoader> getClassLoaders()
    {
      return metaClassLoaders;
    }

    /**
     * Retrieves the MetaClassLoader with the specified Name
     */
    public static MetaClassLoader getClassLoader( String name )
    {
      for( MetaClassLoader mcl : getClassLoaders() )
      {
        if ( name.equals( mcl.getName() )) return mcl;
      }

      return null;
    }

    /**
     * Retrieves the MetaClassLoader for the specified Object
     */
    public static MetaClassLoader findClassLoader( Object obj )
    {
        for( MetaClassLoader l : getClassLoaders() ) {
            if ( l.handles( obj )) return l;
        }

        return null;
    }

    /**
     * Whether the MetaClassLoader handles the object specified
     */
    protected boolean handles( Object obj )
    {
        if ( getMetaClassForObject( obj ) != null ) return true;
        return false;
    }

    /**
     * Retrieves the MetaClass for the specified Object
     */
    public static MetaClass findMetaClass( Object obj )
      throws MetaClassNotFoundException
    {
      // This is a High-Performance addition for MetaObjects
      if ( obj instanceof MetaClassAware ) {
        return ((MetaClassAware) obj ).getMetaClass();
      }

      MetaClassLoader l = findClassLoader( obj );
      if ( l == null )
        throw new MetaClassNotFoundException( "No MetaClass exists for object of class [" + obj.getClass().getName() + "]" );

      return l.getMetaClassForObject( obj );
    }

    /**
     * Retrieves the MetaClass with the specified name
     */
    static MetaClass findMetaClassWithName( String name )
      throws MetaClassNotFoundException
    {
        for( MetaClassLoader l : getClassLoaders() )
        {
            try {
              return l.getMetaClass( name );
            }
            catch( MetaClassNotFoundException e ) {}
        }

        throw new MetaClassNotFoundException( "MetaClass with name [" + name + "] not found" );
    }

    /**
     * Retrieves a collection of all Meta Classes
     */
    public List<MetaClass> getMetaClasses()
    {
    	Collection<MetaData> children = getChildren( MetaClass.class );
    	ArrayList<MetaClass> classes = new ArrayList<MetaClass>( children.size() );
    	for( MetaData md : children ) classes.add( (MetaClass) md );
    	return classes;
    }

    /**
     * Gets the MetaClass of the specified Object
     */
    protected MetaClass getMetaClassForObject( Object obj )
    {
      for( MetaClass mc : getMetaClasses() )
      {
          if ( mc.produces( obj  )) return mc;
      }

      return null;
    }

    /**
     * Gets the MetaClass with the specified name
     */
    public MetaClass getMetaClass( String className ) throws MetaClassNotFoundException
    {
      MetaClass mc = metaClassCache.get( className );
      if ( mc == null )
      {
          synchronized( metaClassCache ) { 
              
            mc = metaClassCache.get( className );
            if ( mc == null )
            {
	      for( MetaClass mc2 : getMetaClasses() )
	      {
	        if ( mc2.getName().equals( className ))
	        {
	          mc = mc2;
	          break;
	        }
	      }

	      if ( mc != null )
	    	  metaClassCache.put( className, mc );
            }
          }
          
          if ( mc == null )
            throw new MetaClassNotFoundException( className ); // "MetaClass with name [" + className + "] not found in MetaClassLoader [" + toString() + "]" );
      }


      return mc;
    }

    /**
     * Adds the MetaClass
     */
    public void addMetaClass( MetaClass mc )
      throws MetaException
    {
      addChild( mc );
    }

    /**
     * Removes the MetaClass
     */
    public void removeMetaClass( String className )
      throws MetaClassNotFoundException
    {
      deleteChild( getMetaClass( className ));
    }

    /**
     * Registers a new MetaClassLoader
     */
    protected static void registerLoader( MetaClassLoader loader )
    {
        metaClassLoaders.add( loader );
    }

    /**
     * Registers a new MetaClassLoader
     */
    protected static void unregisterLoader( MetaClassLoader mcl )
    {
        metaClassLoaders.remove( mcl );
    }

    /**
     * Unloads the MetaClassLoader
     */
    public void destroy()
    {
      log.info( "Destroying the [" + getClass().getSimpleName() + "] MetaClassLoader" );      

      // Remove all classes
      clearChildren();

      // Unregister the class loader
      unregisterLoader( this );      
    }

    ////////////////////////////////////////////////////
    // MISC METHODS

    public String toString()
    {
      if ( getParent() == null )
        return "MetaClassLoader[" + getName() + "]";
      else
        return "MetaClassLoader[" + getName() + "@" + getParent().toString() + "]";
    }
}
