/*
 * Copyright 2002 Draagon Software LLC. All Rights Reserved.
 *
 * This software is the proprietary information of Draagon Software LLC.
 * Use is subject to license terms.
 */

package com.draagon.meta;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.WeakHashMap;

import java.io.Serializable;

import java.lang.ref.WeakReference;
import java.util.Collections;
import java.util.List;
import java.util.Map;

public abstract class MetaData implements Cloneable, Serializable
{
  //private static Log log = LogFactory.getLog( MetaData.class );

  private WeakReference<MetaData> mParentRef = null;
  private String mName = null;
  private List<MetaData> mChildren = null;
  private MetaData mSuperData = null;

  private Map<Object, Object> mCache = null;

  /**
   *  Attribute Options are used to specify expected attributes on MetaData.
   */
  public final class AttributeOption
  {
    final private String name;
    final private Class<?> clazz;
    final private boolean required;
    final private String description;

    /**
     *  Attribute Options are used to specify expected attributes on MetaData.
     *
     * @param name The expected name of the attribute
     * @param type The expected class of the return type of the attribute (use null for any)
     * @param required Whether the attribute is required
     * @param desc A description of the attribute
     */
    public AttributeOption( String name, Class<?> type, boolean required, String desc )
    {
      this.name = name;
      this.clazz = type;
      this.required = required;
      this.description = desc;
    }

    public String getName() { return name; }
    public String getDescription() { return description; }
    public Class<?> getType() { return clazz; }
    public boolean isRequired() { return required; }
  }

  /**
   * Constructs the MetaClass
   */
  public MetaData()
  {
    mChildren = Collections.synchronizedList( new ArrayList<MetaData>() );
    mCache = Collections.synchronizedMap( new WeakHashMap<Object,Object>() );
  }

  public abstract Class<?> getMetaDataClass();

  ////////////////////////////////////////////////////
  // SETTER / GETTER METHODS

  /**
   * Sets the Name of the MetaData
   */
  public void setName( String name )
  {
    mName = name;
  }

  /**
   * Returns the Name of this piece of MetaData
   */
  public String getName()
  {
    return mName;
  }

  /**
   * Sets the parent of the attribute
   */
  protected void attachParent( MetaData parent )
  {
    mParentRef = new WeakReference<MetaData>( parent );
  }

  /**
   * Gets the attribute name
   */
  protected MetaData getParent()
  {
    if ( mParentRef == null ) return null;
    return mParentRef.get();
  }

  /**
   * Sets the Super Data
   */
  public void setSuperData( MetaData superData )
  {
    mSuperData = superData;
  }

  /**
   * Gets the Super Data
   */
  public MetaData getSuperData()
  {
    return mSuperData;
  }

  ////////////////////////////////////////////////////
  // ATTRIBUTE METHODS

  /**
   * Sets an attribute of the MetaClass
   */
  public void addAttribute( MetaAttribute attr )
    throws InvalidAttributeValueException
  {
    addChild( attr );
  }

  /**
   * Sets an attribute of the MetaClass
   */
  public void deleteAttribute( String name )
    throws MetaAttributeNotFoundException
  {
    try {
      deleteChild( name, MetaAttribute.class );
    }
    catch( MetaException e ) {
      throw new MetaAttributeNotFoundException( "MetaAtribute [" + name + "] not found in [" + toInternalString() + "]" );
    }
  }

  /**
   * Sets an attribute value of the MetaData
   */
  public void setAttribute( String name, Object value )
    throws MetaAttributeNotFoundException, InvalidAttributeValueException
  {
    MetaAttribute ma = null;
    try {
      ma = (MetaAttribute) getChild( name, MetaAttribute.class );
    }
    catch( MetaDataNotFoundException e )
    {
      throw new MetaAttributeNotFoundException( "MetaAttribute [" + name + "] was not found in [" + toInternalString() + "]" );

      // If it didn't exist, then create it
      //ma = new MetaAttribute();
      //ma.setName( name );
      //try {
      //  addChild( ma, MetaAttribute.class );
      //} catch( InvalidMetaDataException ex ) {
      //  throw new InvalidAttributeValueException( ex.getMessage(), ex );
      //}
    }

    ma.setValue( value );
  }

  /**
   * Retrieves an attribute value of the MetaData
   */
  public Object getAttribute( String name )
    throws MetaAttributeNotFoundException
  {
    try {
      MetaAttribute ma = (MetaAttribute) getChild( name, MetaAttribute.class );
      return ma.getValue();
    }
    catch( MetaDataNotFoundException e ) {
      throw new MetaAttributeNotFoundException( name ); //"MetaAtribute [" + name + "] not found in [" + toInternalString() + "]" );
    }
  }

  /**
   * Retrieves all attribute names
   */
  public boolean hasAttribute( String name )
  {
    try {
      if ( getChild( name, MetaAttribute.class, false ) != null )
    	  return true;
    }
    catch( MetaDataNotFoundException e ) {}

    return false;
  }

  /**
   * Retrieves all attribute names
   */
  public Collection<MetaAttribute> getAttributes()
  {
	Collection<MetaAttribute> attrs = new ArrayList<MetaAttribute>();
	for ( MetaData md : getChildren( MetaAttribute.class )) {
		attrs.add( (MetaAttribute) md );
	}
    return attrs;
  }

  ////////////////////////////////////////////////////
  // CHILDREN METHODS

  /**
   * Whether the child data exists
   */
  /*public boolean hasChild( String name )
  {
    try {
      getChild( name );
      return true;
    }
    catch( MetaDataNotFoundException e ) {
      return false;
    }
  }*/

  /**
   * Whether the child data exists
   */
  protected boolean hasChild( String name, Class<?> c )
  {
    try {
      getChild( name, c );
      return true;
    }
    catch( MetaDataNotFoundException e ) {
      return false;
    }
  }

  /**
   * Adds a child MetaData object of the specified class type.  If no
   * class type is set, then a child of the same type is not checked against.
   */
  protected void addChild( MetaData data )
    throws InvalidMetaDataException
  {
    addChild( data, true );
  }

  /**
   * Adds a child MetaData object of the specified class type.  If no
   * class type is set, then a child of the same type is not checked against.
   */
  protected void addChild( MetaData data, boolean checkExists )
    throws InvalidMetaDataException
  {
    if ( data == null )
      throw new IllegalArgumentException( "Cannot add null MetaData" );

    if ( checkExists ) {
      try {
        MetaData d = getChild( data.getName(), data.getMetaDataClass() );
        if ( d.getParent() == this )
        {
          if ( d instanceof MetaAttribute
              || d instanceof MetaValidator
              || d instanceof MetaView
              /*|| d instanceof MetaField*/ )
            deleteChild( d );
          else
            throw new InvalidMetaDataException( "MetaData [" + data.toInternalString() + "] with name [" + data.getName() + "] already exists in [" + toInternalString() + "] as [" + d + "]" );
        }
      }
      catch( MetaDataNotFoundException  e ) {}
    }

    data.attachParent( this );
    mChildren.add( data );
  }

  /**
   * Deletes a child MetaData object
   */
  /*public void deleteChild( String name )
    throws MetaException
  {
    deleteChild( name, null );
  }*/

  /**
   * Deletes a child MetaData object of the given class
   */
  protected void deleteChild( String name, Class<?> c )
    throws MetaException
  {
    MetaData d = getChild( name, c );
    if ( d.getParent() == this )
      mChildren.remove( d );
    else
      throw new MetaDataNotFoundException( "You cannot delete MetaData with name [" + name + "] from a SuperData of [" + toInternalString() + "]" );
  }

  /**
   * Deletes a child MetaData object
   */
  public void deleteChild( MetaData data )
  {
    if ( data.getParent() != this )
      throw new IllegalArgumentException( "MetaData [" + data.toInternalString() + "] is not a child of [" + toInternalString() + "]" );

    mChildren.remove( data );
  }

  /*protected MetaData wrapMetaData( MetaData orig )
  {
    //return orig;
    MetaData copy = (MetaData) orig.clone();
    copy.clearChildren();
    copy.setSuperData( orig );
    copy.attachParent( this );
    return copy;
  }*/

  /**
   * Returns all MetaData children
   */
  protected Collection<MetaData> getChildren()
  {
    return getChildren( null );
  }

  /**
   * Returns all MetaData children which implement the specified class
   */
  protected Collection<MetaData> getChildren( Class<?> c )
  {
    // Get all the local children
    ArrayList<MetaData> al = new ArrayList<MetaData>();
    for ( MetaData d : mChildren )
    {
      if ( c == null || c.isInstance( d ))
        al.add( d );
    }

    // Add the super class's children
    if ( getSuperData() != null )
    {
      for( MetaData d : getSuperData().getChildren( c ))
      {
        // We know it exists in the super class, so get it correctly, which adds
        // it to this class except for MetaAttributs
        // if ( tmp instanceof MetaAttribute ) d = tmp;
        // else d = getChild( tmp.getName(), tmp.getMetaDataClass() );

        boolean found = false;

        // Only add the field if it's not found in the super class
        for ( MetaData sd : al )
        {
          if ( sd.getName().equals( d.getName() )) found = true;
        }

        if ( !found ) {
          //if (!( d instanceof MetaAttribute ))
          //  d = getChild( d.getName(), d.getMetaDataClass() );

          al.add( d ); //wrapMetaData( d ));
        }
      }
    }

    return al;
  }

  /**
   * Returns the first child record
   */
  /*public MetaData getFirstChild()
  {
    return getFirstChild( null );
  }*/

  /**
   * Returns the first child record
   */
  protected MetaData getFirstChild( Class<?> c )
  {
    Iterator<MetaData> i = getChildren( c ).iterator();
    if ( !i.hasNext() ) return null;
    return (MetaData) i.next();
  }

  /**
   * Returns a child by the specified name.
   *
   * @param name The name of the child to retrieve.  A null will return the first child.
   */
  /*public MetaData getChild( String name )
    throws MetaDataNotFoundException
  {
    return getChild( name, null );
  }*/

  /**
   * Returns a child by the specified name of the specified class
   *
   * @param name The name of the child to retrieve.  A null will return the first child.
   */
  protected final MetaData getChild( String name, Class<?> c ) throws MetaDataNotFoundException
  {
	  return getChild( name, c, true );
  }

  protected final MetaData getChild( String name, Class<?> c, boolean shouldThrow ) throws MetaDataNotFoundException
  {
    for ( MetaData d : mChildren )
    {
      if (( name == null || d.getName().equals( name ))
        && ( c == null || c.isInstance( d )) ) {
          return d;
      }
    }

    // See if it exists in the parent class
    if ( getSuperData() != null )
    {
      try {
        MetaData md = getSuperData().getChild( name, c, shouldThrow );

        // If the MetaData is not an attribute, then clone, clear,
        //  and set the super data
        /*if ( !( md instanceof MetaAttribute ))
        {
          MetaData add = (MetaData) md.clone();
          add.clearChildren();
          add.setSuperData( md );
          addChild( add, false );
          return add;
        }*/

        return md; //wrapMetaData( md );
      }
      catch( MetaDataNotFoundException ex ) {}
    }

    if ( shouldThrow )
    	throw new MetaDataNotFoundException( name ); // "MetaData child of class [" + c + "] with name [" + name + "] not found in [" + toString() + "]" );
    else
    	return null;
  }

  /**
   * Clears all children
   */
  protected void clearChildren()
  {
    mChildren.clear();
  }

  /**
   * Clears all children of the specified type
   */
  protected void clearChildren( Class<?> c )
  {
    for ( MetaData d : mChildren )
    {
      if ( c == null || c.isInstance( d ))
        mChildren.remove( d );
    }
  }

  ////////////////////////////////////////////////////
  // MISC METHODS

  /**
   *  Returns a list of expected attributes
   */
  public Collection<AttributeOption> getAttributeOptions()
  {
    return new ArrayList<AttributeOption>();
  }

  /**
   * Validates the state of the data in the MetaData object
   */
  public void validate()
    throws InvalidMetaDataException
  {
    if ( getName() == null )
      throw new InvalidMetaDataException( "MetaData has no name" );

    // Validate the Attribute options
    for( AttributeOption ao : getAttributeOptions() )
    {
      boolean has = hasAttribute( ao.getName() );

      // Check for required fields
      if ( ao.isRequired() && !has )
        throw new InvalidMetaDataException( "Required attribute [" + ao.getName() + "] was not found" );

      // Check for proper class types
      if ( has && ao.getType() != null ) {
        try {
          Object obj = getAttribute( ao.getName() );
          if ( !ao.getType().isInstance( obj ))
            throw new InvalidMetaDataException( "Attribute [" + ao.getName() + "] was not an instance of class [" + ao.getType() + "]" );
        } catch( MetaAttributeNotFoundException e ) {}
      }
    }

    // Validate the children
    for( MetaData d : getChildren() ) d.validate();
  }

  public MetaData wrap()
  {
    MetaData d = (MetaData) clone();
    d.clearChildren();
    d.setSuperData( this );
    return d;
  }

  /**
   * Clones this MetaData object
   */
  public Object clone()
  {
    MetaData v;
    try {
      v = (MetaData) this.getClass().newInstance();
    } catch (InstantiationException e) {
      throw new RuntimeException( "Could not create new instance of MetaData class [" + getClass() + "]: " + e.getMessage(), e );
    } catch (IllegalAccessException e) {
      throw new RuntimeException( "Could not create new instance of MetaData class [" + getClass() + "]: " + e.getMessage(), e );
    }

    v.mName = mName;
    v.mParentRef = mParentRef;

    for( MetaData md : getChildren() )
    {
      v.addChild( (MetaData) md.clone() );
      //v.mChildren = (ArrayList) mChildren.clone();
    }

    return v;
  }

  /**
   * Sets a cache value for this piece of MetaData
   */
  public void setCacheValue( Object key, Object value )
  {
	  //ystem.out.println( "SET [" + key + "] = " + value );
	  mCache.put( key, value );
  }

  /**
   * Retrieves a cache value for this piece of MetaData
   */
  public Object getCacheValue( Object key )
  {
	  return mCache.get( key );
  }

  /**
   * Returns a string representation of the MetaData
   */
  private String toInternalString()
  {
    String name = getClass().getName();
    int i = name.lastIndexOf( '.' );
    if ( i >= 0 ) name = name.substring( i + 1 );

    if ( getParent() == null )
      return name + "{" + getName() + "}";
    else
      return name + "{" + getName() + "}@" + getParent().toInternalString();
  }

  /**
   * Returns a string representation of the MetaData
   */
  public String toString()
  {
    String name = getClass().getName();
    int i = name.lastIndexOf( '.' );
    if ( i >= 0 ) name = name.substring( i + 1 );

    if ( getParent() == null )
      return name + "{" + getName() + "}";
    else
      return name + "{" + getName() + "}@" + getParent().toString();
  }
}


